﻿/*****************************************************
 * Map.cs
 * ラスタ形式の地図情報を扱うことを念頭に入れた基底クラスです。
 * 
 * マップの取り扱い注意
 * 1)本クラスのコンストラクタ生成時には必ずマップの右上と左下の座標（緯度・経度）をセットして下さい。
 * 
 * 　開発者：K.Morishita Kumamoto-University Japan
 * 　動作環境：.NET Framework 4
 * 　開発履歴：
 * 　        2011/7/2    NA値の追加
 * 　                    相対座標から緯度・経度に直すなどのメソッドを整備
 * 　        2011/7/6    Position2MeshCorner等関連メソッドを定義
 * 　                    DEMデータでは標高0m以下を-9999.0で表すので、NA値を-99999.0に変更した。
 * 　        2011/9/16   Initialize()にデフォルト値を設定した
 * 　        2011/10/2   メモ：将来、+演算子に対応させて重ね合わせをやってみたいと思う。
 * 　                    今後の用途によるが、nameや属性・単位等がメンバにあっても良いと思う。
 * 　        2011/10/3   メモ：マップ同士の重ね合わせがあるといいなぁと思う。
 * 　        2011/11/20  LocalData構造体にToString()を追加
 * 　        2011/12/15  MapBase内のメッシュサイズ計算における引数の渡す順番ミスを修正
 * 　                    Pos構造体にCompareTo()を追加した。Array.Sort()で配列をソートできるようになった。
 * 　                    MapLocalFeatureクラスを新設
 * 　                    GetMapBetween(),GetMapHiger(),GetMapLower(),GetMapMatch(),GetMatchPosArray(),SaveAsImage()を整備。正常な動作を確認した。
 * 　        2011/12/16  コメントの見直しと関数の整理などを実施した。
 * 　                    マップ内のローカルな座標系の配列を緯度・経度の配列に変換するメソッドToBLHarray()を実装した。
 * 　        2011/12/18  MapCorner構造体を、コンストラクタ生成時だけでなく要素にアクセスする際にも引数をチェックするようにプロパティを新設
 * 　                    今までpublicにしていた座標値はprivateにして隠ぺいした。
 * 　                    おかげで他のソースコードでの変更点が多数あったが、エラー原因の追及時間が短くなると思う。
 * 　                    あと、MapCorner構造体にToString()を追加した。
 * 　                     
 * 　                    GetValue(Pos _pos)を整備し、GetValue(BLH _position)の構造をそれに併せて変更した。
 * 　                    GetPosDiff()をGetDifferentialPosition()に名称変更
 * 　                    Pos[]配列によるAddValue()を新設
 * 　                    PosのUseful()メソッドを新設
 * 　        2012/4/30   PosへEquals()を追加
 * 　                    MapLocalFeatureクラスを本ソースファイルより分離
 * 　        2012/5/4    クラス内部に宣言している構造体を別ファイルに移した。
 * 　        2012/5/28   Map.NAを廃止し、該当データが存在しない場合にはfloat.NaNを返すように仕様を変更した。
 * 　                    DEMクラスのマップ生成アルゴリズムの大幅な変更に伴い利用しなくなったメソッドを削除した。
 * 　        2012/5/29   SaveAsImage()を高速化
 * 　        2012/6/2    メッシュ数を表すプロパティ名がLenghになっていたのをLengthと訂正した。
 * 　        2012/6/4    コピーコンストラクタを新設した。
 * 　                    これでマップのインスタンスを安全にコピーできるようになった。
 * 　        2012/6/8    一部のコメントがコードと一致していなかったため変更した。
 * 　                    また、メソッド名に使っていたXMLをXmlへと変更した。
 * ***************************************************/
using System;
using System.IO;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Drawing;
using System.Threading;
using System.Threading.Tasks;
using System.Drawing.Imaging;
using System.Runtime.InteropServices;
using GNSS;
using GNSS.Field;
using Graphic.ColorMap;
using Graphic;
 
namespace GIS.MapBase
{
    /// <summary>
    /// ラスタデータで構成される地図情報を扱うクラス
    /// <para>DEM（デジタル標高モデル）等を扱うことを念頭に入れています。</para>
    /// <para>
    /// 要素を2次元配列として表現しています。
    /// 配列は[x,y]で宣言しており、画像でたとえると[0,0]が画面の左上、[x_max, y_max]が画面の右下としています。
    /// 画面の上が北で、右が東です。
    /// 東西方向をx、南北方向をyで表現しています。
    /// xは増えるほどより東に位置することを表し、yは増えるほどより南に位置することを表します。
    /// </para>
    /// </summary>
    public partial class Map
    {
        #region /*　メンバ変数　***************************************************************/
        /// <summary>
        /// マップの領域
        /// </summary>
        private RectangleField field;
        /// <summary>
        /// マップのサイズ
        /// </summary>
        private Pos size;
        /// <summary>
        /// マップサイズに、newとSetMap間で不一致があるとtrue
        /// </summary>
        private Boolean errorSize = false;
        /// <summary>
        /// メッシュの平均的サイズ
        /// <para>緯度方向・経度方向のメッシュサイズとする。単位は[m]。</para>
        /// </summary>
        private Length meshSise = new Length();
        /// <summary>
        /// ラスタのマップデータ
        /// </summary>
        private float[,] d_map;
        /// <summary>
        /// 画像として保存する時のフォーマット
        /// </summary>
        private ImageFormat imageFormat;
        #endregion
        #region /*　プロパティ　***************************************************************/
        /// <summary>
        /// マップのサイズ
        /// <para>メッシュの数でカウントしたマップのサイズを返します。</para>
        /// </summary>
        public Pos Size 
        { 
            get {
                return this.size; 
            }
            protected set {
                this.size = value;
            }
        }
        /// <summary>
        /// マップがセット済みかどうかをtrue or falseで返す。
        /// <para>セットされていればtrue。</para>
        /// </summary>
        public Boolean IsMapSet
        {
            get { 
                if(this.d_map != null)
                    return true;
                else
                    return false;
            }
        }
        /// <summary>
        /// マップのサイズがセットされているかを確認する。
        /// セットされていればtrue。
        /// </summary>
        public Boolean IsSizeSet
        {
            get
            {
                Boolean ans = false;

                if (!(this.Size.x == 0 || this.Size.y == 0)) ans = true;    // マップが2次元として大きさを持っていればtrue
                return ans;
            }
        }
        /// <summary>
        /// メッシュサイズ
        /// <para>南北方向・東西方向のメッシュの平均的サイズ[m]です。</para>
        /// </summary>
        public Length MeshSize 
        { 
            get { return this.meshSise; }
            protected set 
            {
                this.meshSise = value;
            }
        }
        /// <summary>
        /// newでセットしたマップサイズとSetMapでセットされたサイズ間に差があるとtrueとなる。
        /// </summary>
        public Boolean IsErrorSize 
        {
            get { 
                return this.errorSize; 
            }
        }
        /// <summary>
        /// マップの座標（緯度・経度）
        /// </summary>
        public RectangleField Field 
        { 
            set {
                if (this.IsSizeSet)                                         // 可能ならメッシュサイズを計算する
                {
                    this.MeshSize = this.GetMeshSize(value, this.Size);
                }
                this.field = value;
            }
            get { return this.field; }
        }
        /// <summary>
        /// マップのメッシュ数
        /// </summary>
        public int Length 
        { 
            get { return this.d_map.GetLength(0) * this.d_map.GetLength(1); }
        }
        /// <summary>
        /// マップ中の最大値を取得する
        /// <para>走査するので少し時間がかかります。</para>
        /// </summary>
        public float Max 
        {
            get {
                int x_width = this.d_map.GetLength(0);
                int y_width = this.d_map.GetLength(1);
                float max = float.MinValue;

                for (int x = 0; x < x_width; x++)
                    for (int y = 0; y < y_width; y++)
                        if (this.d_map[x, y] > max) max = this.d_map[x, y];
                return max;
            }
        }
        /// <summary>
        /// マップ中の最大値を取得する
        /// <para>走査するので少し時間がかかります。</para>
        /// </summary>
        public float Min
        {
            get
            {
                int x_width = this.d_map.GetLength(0);
                int y_width = this.d_map.GetLength(1);
                float min = float.MaxValue;

                for (int x = 0; x < x_width; x++)
                {
                    for (int y = 0; y < y_width; y++)
                    {
                        if (this.d_map[x, y] < min && this.d_map[x, y] != float.NaN) min = this.d_map[x, y];
                    }
                }
                return min;
            }
        }
        /// <summary>
        /// マップを画像出力するときのフォーマット
        /// </summary>
        public System.Drawing.Imaging.ImageFormat ImageFormat
        {
            get { return this.imageFormat; }
            set { this.imageFormat = value; }
        }
        /// <summary>
        /// メッシュのアスペクト比
        /// <para>N / E で定義している。つまり、1以上なら南北方向に長いメッシュである。</para>
        /// </summary>
        public double MeshAspectRatio 
        { 
            get 
            { 
                return this.MeshSize.N / this.MeshSize.E; 
            }
        }
        /// <summary>
        /// 本マップの中央の座標
        /// </summary>
        public GNSS.BLH Center
        {
            get { return this.Field.UpperRight.GetMedian(this.Field.LowerLeft); }
        }
        #endregion
        /*　インデクサ　*********************************************************************/
        /// <summary>
        /// 指定座標の値（例えば高度）を取得orセット
        /// <para>座標はマップのローカル座標（0～this.Size.x or y - 1）</para>
        /// <para>[0 ,0]がマップの左上を指し、xは右向きに正，yは下向きに正となります。</para>
        /// </summary>
        /// <param name="x">x座標</param>
        /// <param name="y">y座標</param>
        /// <returns>マップの値</returns>
        public float this[int x, int y]
        {
            get {
                if (x >= 0 && y >= 0 && x < this.Size.x && y < this.Size.y)
                    return this.d_map[x, y];
                else
                    throw new SystemException("引数が範囲内にありません。");
            }
            set { 
                if (x >= 0 && y >= 0 && x < this.Size.x && y < this.Size.y)
                    this.d_map[x, y] = value;
                else
                    throw new SystemException("引数が範囲内にありません。");
            }
        }
        #region /*　メンバ関数（メソッド）***************************************************/
        /// <summary>
        /// 指定値よりも低い値を持つ地点をマーキングしたマップを返す
        /// <para>返されたマップは、1.0を条件成立とする。条件不成立は0.0とする。</para>
        /// </summary>
        /// <param name="value">閾値</param>
        /// <returns>マップ</returns>
        public Map GetMapLower(float value)
        {
            float[,] newMap = new float[this.d_map.GetLength(0), this.d_map.GetLength(1)];
            int x_width = this.Size.x;
            int y_height = this.Size.y;

            Parallel.For(0, x_width, x =>                                       // 配列に格納
            {
                Parallel.For(0, y_height, y =>
                {
                    if (this.d_map[x, y] < value)
                        newMap[x, y] = (float)1.0;
                    else
                        newMap[x, y] = (float)0.0;
                });
            });

            Map map = new Map(this.Field);
            map.SetMapData(newMap);
            return map;
        }
        /// <summary>
        /// 指定値よりも高い値を持つ地点をマーキングしたマップを返す
        /// <para>返されたマップは、1.0を条件成立とする。条件不成立は0.0とする。</para>
        /// </summary>
        /// <param name="value">閾値</param>
        /// <returns>マップ</returns>
        public Map GetMapHigher(float value)
        {
            float[,] newMap = new float[this.d_map.GetLength(0), this.d_map.GetLength(1)];
            int x_width = this.Size.x;
            int y_height = this.Size.y;

            Parallel.For(0, x_width, x =>                                       // 配列に格納
            {
                Parallel.For(0, y_height, y =>
                {
                    if (this.d_map[x, y] > value)
                        newMap[x, y] = (float)1.0;
                    else
                        newMap[x, y] = (float)0.0;
                });
            });

            Map map = new Map(this.Field);
            map.SetMapData(newMap);
            return map;
        }
        /// <summary>
        /// 指定値の間にある値を持つ地点をマーキングしたマップを返す
        /// <para>なお、条件には指定値は含まれない。</para>
        /// <para>2つの引数が同値であれば無評価となり、0で初期化されたマップが返ります。</para>
        /// </summary>
        /// <param name="value1">値1</param>
        /// <param name="value2">値2</param>
        /// <returns>マップ</returns>
        public Map GetMapBetween(float value1, float value2)
        {
            float[,] newMap = new float[this.d_map.GetLength(0), this.d_map.GetLength(1)];
            int x_width = this.Size.x;
            int y_height = this.Size.y;

            if (value1 != value2)
            {
                float upper, lower;
                if (value1 > value2)
                {
                    upper = value1;
                    lower = value2;
                }
                else
                {
                    upper = value2;
                    lower = value1;
                }
                Parallel.For(0, x_width, x =>                                       // 配列に格納
                {
                    Parallel.For(0, y_height, y =>
                    {
                        if (this.d_map[x, y] > lower && this.d_map[x, y] < upper)
                            newMap[x, y] = (float)1.0;
                        else
                            newMap[x, y] = (float)0.0;
                    });
                });
            }
            Map map = new Map(this.Field);
            map.SetMapData(newMap);
            return map;
        }
        /// <summary>
        /// 指定値にマッチする地点をマーキングしたマップを返す
        /// <para>指定座標の周囲を考慮して、指定値が指定座標の持つ値と合致するかをチェックし、合致した地点を1.0としたマップを返します。</para>
        /// <para>走査範囲については、fieldで指定して下さい。</para>
        /// </summary>
        /// <param name="value">閾値</param>
        /// <param name="field">マップ中の走査範囲領域</param>
        /// <returns>マップ</returns>
        public Map GetMapMatch(float value, MapField field)
        {
            field = field.And(new MapField(this.Size));                                         // 領域を本インスタンスに合わせて調整（領域が縮小することも有り得る）
            float[,] newMap = new float[this.d_map.GetLength(0), this.d_map.GetLength(1)];
            int x_limit = field.LowerRight.x + 1;
            int y_limit = field.LowerRight.y + 1;
            MapLocalFeature feature = new MapLocalFeature(this);                                // 特徴を計算してくれるクラスを宣言

            Parallel.For(field.UpperLeft.x, x_limit, x =>                                       // 配列に格納
            {
                Parallel.For(field.UpperLeft.y, y_limit, y =>
                {
                    if (feature.GetFeature(x, y).CheckBetween(value))
                        newMap[x, y] = (float)1.0;
                    else
                        newMap[x, y] = (float)0.0;
                });
            });

            Map map = new Map(this.Field);                                                      // マップの座標をセット
            map.SetMapData(newMap);                                                             // マップの中身をセット
            return map;
        }
        /// <summary>
        /// 指定値にマッチする地点をマーキングしたマップを返す
        /// <para>指定座標の周囲を考慮して、指定値が指定座標の持つ値と合致するかをチェックし、合致した地点を1.0としたマップを返します。</para>
        /// <para>走査範囲はマップ全体となります。</para>
        /// </summary>
        /// <param name="value">閾値</param>
        /// <returns>マップ</returns>
        public Map GetMapMatch(float value)
        {
            return this.GetMapMatch(value, new MapField(new Pos(), this.Size));
        }
        /// <summary>
        /// 指定値にマッチする地点の座標（マップのローカル座標）を配列で返します。
        /// <para>GetMapMatchを使用して、フラグの立った地点を配列として返します。</para>
        /// <para>走査範囲については、fieldで指定して下さい。</para>
        /// </summary>
        /// <param name="value">閾値</param>
        /// <param name="field">マップ中の走査範囲領域</param>
        /// <returns>マッチした地点の座標を配列にしたもの</returns>
        public Pos[] GetMatchPosArray(float value, MapField field)
        {
            field = field.And(new MapField(this.Size));                                         // 領域を本インスタンスに合わせて調整（領域が縮小することも有り得る）
            List<Pos> ans = new List<Pos>(0);
            int x_limit = field.LowerRight.x + 1;
            int y_limit = field.LowerRight.y + 1;
            Map matchMap = this.GetMapMatch(value, field);

            Parallel.For(field.UpperLeft.x, x_limit, x =>                                       // Listに格納
            {
                Parallel.For(field.UpperLeft.y, y_limit, y =>
                {
                    if (matchMap[x, y] != 0.0)
                    {
                        Monitor.Enter(ans);                                                     // 通常はフラグの立っている件数の方が圧倒的に少ないはずなので、普通のfor文よりもこの方が速いはず。
                        ans.Add(new Pos(x, y));
                        Monitor.Exit(ans);
                    }
                });
            });
            return ans.ToArray();
        }
        /// <summary>
        /// 指定値にマッチする地点の座標（マップのローカル座標）を配列で返します。
        /// <para>GetMapMatchを使用して、フラグの立った地点を配列として返します。</para>
        /// <para>走査範囲はマップ全体となります。</para>
        /// </summary>
        /// <param name="value">閾値</param>
        /// <returns>マッチした地点の座標を配列にしたもの</returns>
        public Pos[] GetMatchPosArray(float value)
        {
            return this.GetMatchPosArray(value, new MapField(new Pos(), this.Size));
        }
        /// <summary>
        /// 指定された座標を中心とした、指定された半径 mを収めることができる領域を返す
        /// <para>返す領域はマップの領域から外に出ることも有り得ます。</para>
        /// </summary>
        /// <param name="center">中心座標</param>
        /// <param name="radius">半径[m]</param>
        /// <returns>指定された範囲が収まるMapField</returns>
        public MapField GetField(Pos center, double radius)
        {
            int delta_x = (int)(radius / this.MeshSize.E);
            int delta_y = (int)(radius / this.MeshSize.N);
            Pos delta = new Pos(delta_x, delta_y);
            return new MapField(center + delta, center - delta);
        }
        /// <summary>
        /// マップを2次元配列で返す
        /// <para>
        /// インスタンスは新たに生成しないので、高速アクセス・操作が可能です。
        /// ただし、値をいじるとマップも変わるので注意して下さい。</para>
        /// </summary>
        public float[,] GetMap()
        {
            return this.d_map;
        }
        /// <summary>
        /// マップを画像として保存する
        /// <para>保存形式はImageFormatプロパティを通して設定可能です。</para>
        /// <para>
        /// 個人的都合により、値が非値もしくは0.0以下の箇所では背景として黒色を採用しています。
        /// これは使いにくする仕様になっていますので、その内カラ―スケールを導入するかもしれません。
        /// </para>
        /// <para>
        /// ファイルの出力には実時間で通常2秒以上かかりますのでTaskを利用して処理がすぐ戻るようにしています。
        /// 処理が戻った時点ではファイルができていないことにご注意ください。
        /// </para>
        /// </summary>
        /// <param name="fname">ファイル名</param>
        /// <returns>保存したファイルのフルパス</returns>
        public string SaveAsImage(string fname)
        {
            string cd = Directory.GetCurrentDirectory();                        // カレントディレクトリ（作業フォルダ）を取得
            string fileName = cd + @"\" + fname;
            Colormap myColorMap = new Colormap();
            int x_width = this.Size.x;
            int y_height = this.Size.y;
            float max = this.Max;
            float min = this.Min;
            int resized_y = (int)(this.MeshAspectRatio * (double)y_height);     // アスペクト比を考慮して、高さを調整する

            Bitmap bitmap;// = new Bitmap(x_width, resized_y);                  // bitmap作成のためのBitmapオブジェクト作成

            /* // 旧コードを念のため残しておく
            Parallel.For(0, x_width, x =>                        
            {                                                                   // 配列の値をbitmapオブジェクトのピクセル値に代入
                Parallel.For(0, resized_y, y =>
                {
                    int origin_y = y_height * y / resized_y;
                    double scale = (this.d_map[x, origin_y] - min) / (max - min);
                    if (scale < 0.0 || double.IsNaN(scale))
                        scale = 0.0; 
                    else if (scale > 1.0) 
                        scale = 1.0;
                    Color color;
                    if (scale == 0.0 || this.d_map[x, origin_y] <= 0.0)
                        color = Color.Black;                                    // 背景は黒
                    else
                        color = myColorMap.getColor(scale);
                    Monitor.Enter(bitmap);                                      // ロックしないとエラーが起きた。。
                    bitmap.SetPixel(x, y, color);
                    Monitor.Exit(bitmap);
                });
            });
            */

            var tsk = Task.Factory.StartNew(() =>
            {
                Color[,] img = new Color[x_width, resized_y];
                Parallel.For(0, x_width, x =>                                       // 配列に色を代入
                {
                    Parallel.For(0, resized_y, y =>
                    {
                        int origin_y = y_height * y / resized_y;
                        double scale = (this.d_map[x, origin_y] - min) / (max - min);
                        if (scale < 0.0 || double.IsNaN(scale))
                            scale = 0.0;
                        else if (scale > 1.0)
                            scale = 1.0;
                        Color color;
                        if (scale == 0.0 || this.d_map[x, origin_y] <= 0.0)
                            color = Color.Black;                                    // 背景は黒
                        else
                            color = myColorMap.getColor(scale);
                        img[x, y] = color;
                    });
                });

                BmpGenerator ge = new BmpGenerator();                               // 作った配列情報を用いて画像を作成する
                ge.MakeBmp(img);
                bitmap = ge.Bitmap;
                bitmap.Save(fileName, this.ImageFormat);                            // bitmapオブジェクトのデータをbmpファイルに保存
            });
            return fileName;
        }
        /*2011/7/6 動作確認済み*/
        /// <summary>
        /// 本マップの情報を文字列にして返す。
        /// <para>
        /// 確認用です。
        /// 1要素毎に改行しています。
        /// 配列への格納順はマップの左上から右の方へ進めて右端で左端に返す様にしています。
        /// </para>
        /// </summary>
        public override string ToString()
        {
            StringBuilder sb = new StringBuilder(13375000);                      // 予め大きなメモリ容量を確保しておく

            if (this.IsMapSet)
            {
                for (int y = 0; y < this.Size.y; y++)
                {
                    for (int x = 0; x < this.Size.x; x++)
                    {
                        sb.Append(this.d_map[x, y].ToString()).Append("\n");    // stringを使うより、断然高速
                    }
                }
            }
            return sb.ToString();
        }
        /// <summary>
        /// 本マップの情報を、XMLフォーマットの文字列にして返す。
        /// <para>
        /// マップとしてはデフォルト形式です。
        /// 確認用です。
        /// </para>
        /// </summary>
        public string ToStringAsXml()
        {
            StringBuilder sb = new StringBuilder(3375000);                      // 予め大きなメモリ容量を確保しておく

            if (this.IsMapSet)
            {
                sb.Append("<gml:high>").Append((this.Size.x - 1).ToString()).Append(" ").Append((this.Size.y - 1).ToString()).Append("</gml:high>\n<gml:tupleList>\n");
                for (int y = 0; y < this.Size.y; y++)
                {
                    for (int x = 0; x < this.Size.x; x++)
                    {
                        sb.Append("other,").Append(this.d_map[x, y].ToString()).Append("\n");    // stringを使うより、断然高速
                    }
                }
                sb.Append("</gml:tupleList>\n");
            }
            return sb.ToString();
        }
        /// <summary>
        /// 本マップの情報を、Rコマンダー用フォーマットの文字列にして返す
        /// <para>
        /// 緯度・経度とマップ情報をカンマ区切りで出力します。
        /// 確認用です。
        /// </para>
        /// </summary>
        public string ToStringForRcmdr()
        {
            StringBuilder sb = new StringBuilder(3375000);                      // 予め大きなメモリ容量を確保しておく

            if (this.IsMapSet)
            {
                sb.Append("Longitude,Latitude,Value\n");
                for (int y = 0; y < this.Size.y; y++)
                {
                    for (int x = 0; x < this.Size.x; x++)
                    {
                        Pos add = new Pos(x, y);
                        BLH pos = this.XYaddress2Position(add);
                        sb.Append(pos.L.ToString()).Append(",").Append(pos.B.ToString()).Append(",").Append(this.d_map[x, y].ToString()).Append("\n");    // stringを使うより、断然高速
                    }
                }
            }
            return sb.ToString();
        }
        /// <summary>
        /// 指定ファイル名でデータを保存する
        /// <para>KMLに加工するメソッドもその内作りたいなぁ。</para>
        /// </summary>
        /// <param name="fname">ファイル名</param>
        public void SaveFileAsXml(string fname)
        {
            using (System.IO.StreamWriter fwriter = new System.IO.StreamWriter(fname, false))
            {
                try
                {
                    fwriter.Write(this.ToStringAsXml());                              // 文字列化したデータを保存する。
                }
                catch 
                {
                    // エラー処理があれば記述
                }
            }
            return;
        }
        /// <summary>
        /// 指定ファイル名でデータを保存する
        /// <para>KMLに加工するメソッドもその内作りたいなぁ。</para>
        /// </summary>
        /// <param name="fname">ファイル名</param>
        public void SaveFileForRcmdr(string fname)
        {
            using (System.IO.StreamWriter fwriter = new System.IO.StreamWriter(fname, false))
            {
                try
                {
                    fwriter.Write(this.ToStringForRcmdr());                              // 文字列化したデータを保存する。
                }
                catch
                {
                    // エラー処理があれば記述
                }
            }
            return;
        }
        /*2011/7/6 動作確認済み*/
        /// <summary>
        /// 渡された緯度経度がこのマップの範囲内にあるかどうかをBoolで返す。
        /// </summary>
        /// <param name="_position">検査座標</param>
        /// <returns>trueなら使用可能</returns>
        public Boolean Useful(BLH _position)
        {
            if (_position.B <= this.Field.UpperRight.B &&
                _position.B >= this.Field.LowerLeft.B &&
                _position.L <= this.Field.UpperRight.L &&
                _position.L >= this.Field.LowerLeft.L)
            {
                return true;
            }else{
                return false;
            }
        }
        /// <summary>
        /// 渡されたローカル座標がこのマップの範囲内にあるかどうかをBoolで返す。
        /// </summary>
        /// <param name="_position">検査座標</param>
        /// <returns>trueなら使用可能</returns>
        /// <exception cref="SystemException">マップサイズがセットされていなければスロー</exception>
        public Boolean Useful(Pos _position)
        {
            if (this.IsSizeSet)
            {
                if (_position.x >= 0 && _position.x < this.Size.x && _position.y >= 0 && _position.y < this.Size.y)
                {
                    return true;
                }
                else
                {
                    return false;
                }
            }
            else
            {
                throw new SystemException("マップサイズがセットされていません。初期化が必要です。");
            }
        }
        /*2011/7/6 動作確認済み*/
        /// <summary>
        /// 指定座標のx,yアドレスを返す
        /// <para>
        /// マップの要素番号は0～size-1であることに注意してください。
        /// 本メソッドはマップ領域外の座標を指定された場合でも本マップ内の相対アドレスを返します。
        /// </para>
        /// </summary>
        /// <param name="_position">指定座標（緯度・経度）</param>
        /// <returns>x,yアドレス</returns>
        public Pos Position2XYaddress(BLH _position)
        {
            Pos ans = new Pos();

            if (this.IsSizeSet)                             // マップサイズがセットしてあることを確認（マップサイズが不明だと内部アドレスを計算できない）
            {
                ans.y = (int)((double)this.Size.y / (this.Field.LowerLeft.B - this.Field.UpperRight.B) * (_position.B - this.Field.UpperRight.B));       // 将来、C#のキャストの仕様が変更になったらバグとなるので注意(2011/7)
                ans.x = (int)((double)this.Size.x / (this.Field.UpperRight.L - this.Field.LowerLeft.L) * (_position.L - this.Field.LowerLeft.L));
            }
            else 
            {
                throw new SystemException("マップにサイズが定義されていません。初期化が必要です。");
            }
            return ans;
        }
        /*2011/7/6 動作確認済み*/
        /// <summary>
        /// 指定アドレスにおけるメッシュ中央の緯度・経度を返す
        /// <para>指定アドレスは、マップの範囲を超えていてもOK。</para>
        /// </summary>
        /// <param name="_add">x,yで指定される座標</param>
        /// <returns>座標（緯度・経度）</returns>
        private BLH XYaddress2Position(Pos _add)
        {
            BLH ans = new BLH();

            if (this.IsSizeSet)                             // マップサイズがセットしてあることを確認（マップサイズが不明だと座標を計算できない）
            {
                ans.B = (this.Field.LowerLeft.B - this.Field.UpperRight.B) / (double)this.Size.y * ((double)_add.y + 0.5d) + this.Field.UpperRight.B;
                ans.L = (this.Field.UpperRight.L - this.Field.LowerLeft.L) / (double)this.Size.x * ((double)_add.x + 0.5d) + this.Field.LowerLeft.L;
            }
            return ans;
        }
        /// <summary>
        /// ローカルアドレスの配列を緯度・経度の配列へ変換して返す
        /// </summary>
        /// <param name="posArray">ローカルアドレスの配列</param>
        /// <returns>緯度・経度に変換した配列</returns>
        public GNSS.BLH[] ToBLHarray(Pos[] posArray)
        {
            GNSS.BLH[] blhArray = new BLH[posArray.Length];
            for (int i = 0; i < blhArray.Length; i++) blhArray[i] = this.XYaddress2Position(posArray[i]);
            return blhArray;
        }
        /// <summary>
        /// 座標（緯度・経度）に指定された値を加算する
        /// <para>加算値は負数でも構いません。</para>
        /// </summary>
        /// <param name="_position">指定座標（緯度・経度）</param>
        /// <param name="value">加算値</param>
        public void AddValue(BLH _position, float value)
        {
            Pos address = new Pos();

            if (this.IsMapSet && this.Useful(_position))            // マップがセットしてあり、且つ指定座標がマップの範囲内であればOK 
            {
                address = this.Position2XYaddress(_position);       // アドレスを計算
                if (address.x >= 0 && address.x < this.Size.x && address.y >= 0 && address.y < this.Size.y)
                    this.d_map[address.x, address.y] += value;      // アドレスが有効範囲であれば、値を加算する
            }
            return;
        }
        /// <summary>
        /// 座標（マップ内のローカル座標）に指定された値を加算する
        /// <para>加算値は負数でも構いません。</para>
        /// </summary>
        /// <param name="_position">指定座標（マップ内のローカル座標）</param>
        /// <param name="value">加算値</param>
        public void AddValue(Pos _position, float value)
        {
            if (this.IsMapSet && this.Useful(_position))            // マップがセットしてあり、且つ指定座標がマップの範囲内であればOK 
            {
                this.d_map[_position.x, _position.y] += value;      // アドレスが有効範囲であれば、値を加算する
            }
            else 
            {
                throw new SystemException("マップが初期化されておりません。");
            }
            return;
        }
        /// <summary>
        /// 座標（緯度・経度の配列）に指定された値（配列）を加算する
        /// </summary>
        /// <param name="_position">指定座標配列（緯度・経度）</param>
        /// <param name="values">加算値の配列（座標毎に個別に指定可能です）</param>
        public void AddValues(BLH[] _position, float[] values)
        {
            Parallel.For(0, _position.Length, i =>
            {
                this.AddValue(_position[i], values[i]);
            });
            return;
        }
        /// <summary>
        /// 座標（マップ内のローカル座標）に指定された値（配列）を加算する
        /// </summary>
        /// <param name="_position">指定座標配列</param>
        /// <param name="values">加算値の配列（座標毎に個別に指定可能です）</param>
        public void AddValues(Pos[] _position, float[] values)
        {
            Parallel.For(0, _position.Length, i =>
            {
                this.AddValue(_position[i], values[i]);
            });
            return;
        }
        /*2011/7/6 動作確認済み*/
        /// <summary>
        /// 指定座標の属するメッシュの右上・左上の座標を返す
        /// <para>任意に作出されたマップの位置を本クラスのメッシュ位置と一致させる用途での仕様を想定しています。</para>
        /// </summary>
        /// <param name="_position">指定座標（緯度・経度）</param>
        /// <returns>メッシュの右上・左上の座標（緯度・経度）</returns>
        public RectangleField GetMeshCorner(BLH _position)
        {
            RectangleField ans = new RectangleField();
            if (this.IsSizeSet)                                 // マップサイズがセットしてあることを確認（マップサイズが不明だと内部アドレスを計算できない）
            {
                Pos add = this.Position2XYaddress(_position);
                BLH centerPos = this.XYaddress2Position(add);   // 指定メッシュの中央位置における緯度・経度を取得
                BLH deltaPos = new BLH();
                deltaPos.B = (this.Field.UpperRight.B - this.Field.LowerLeft.B) / ((double)this.Size.y * 2.0);   // メッシュ中央から、メッシュ境界までの距離を緯度・経度で計算
                deltaPos.L = (this.Field.UpperRight.L - this.Field.LowerLeft.L) / ((double)this.Size.x * 2.0);
                BLH upperR = centerPos + deltaPos;              // 右上の座標を計算
                BLH underL = centerPos - deltaPos;              // 左下の座標を計算
                ans = new RectangleField(upperR, underL);
            }
            return ans;
        }
        /*2011/7/6 動作確認済み*/
        /// <summary>
        /// 指定された座標を中心として、指定された相対アドレスの緯度と経度を返します。
        /// </summary>
        /// <param name="_center_position">指定座標</param>
        /// <param name="_delta_address">相対座標（メッシュ数でカウントする）</param>
        /// <returns>指定された座標のマップ情報を返す</returns>
        public BLH GetDifferentialPosition(BLH _center_position, Pos _delta_address)
        {
            BLH ans = new BLH();

            if (this.IsSizeSet)                             // マップサイズがセットしてあることを確認（マップサイズが不明だと内部アドレスを計算できない）
            {
                Pos current_pos = this.Position2XYaddress(_center_position);
                current_pos += _delta_address;              // 指定されている相対座標を加算する
                ans = this.XYaddress2Position(current_pos); // 緯度・経度を求める
            }
            return ans;
        }
        /// <summary>
        /// 2地点間の距離[m]を返す
        /// </summary>
        /// <param name="p1">地点1</param>
        /// <param name="p2">地点2</param>
        /// <returns>2地点間の距離[m]</returns>
        public double GetDistance(Pos p1, Pos p2)
        {
            Pos diff = p1 - p2;
            return Math.Sqrt(Math.Pow(diff.x * this.MeshSize.E, 2.0) + Math.Pow(diff.y * this.MeshSize.N, 2.0));
        }
        /*2011/7/6 動作確認済み*/
        /// <summary>
        /// 指定座標のマップデータを返す
        /// <para>
        /// 計算式の関係上、東経を基準にしているので西経のデータを扱う場合は東経に変換してから使用する事。
        /// 同様に、南緯は北緯に変換して使用して下さい。
        /// 例：西経15度=>東経-15度
        /// </para>
        /// </summary>
        /// <param name="_position">座標（lat,lon）</param>
        /// <returns>指定座標のマップデータ。ただし、該当なしの場合はfloat.NaNを返す。</returns>
        public float GetValue(BLH _position)
        {
            float ans = float.NaN;

            if (this.IsMapSet)                                                      // マップがセットされていなければ、渡すデータが無い（それに、エラーが出る）
            {
                if (this.Useful(_position))                                         // 指定座標が本マップに適用可能なら、値を抽出する
                {
                    Pos address = this.Position2XYaddress(_position);               // 指定座標からマップ内アドレスを計算
                    ans = this.GetValue(address);
                }
            }
            else
            {
                throw new SystemException("マップがセットされていません。初期化が必要です。");
            }
            return ans;
        }
        /// <summary>
        /// 指定座標のマップデータを返す
        /// <para>
        /// マップ内のローカル座標を指定して下さい。
        /// </para>
        /// </summary>
        /// <param name="_pos">マップ内のローカル座標</param>
        /// <returns>指定座標があれば指定されたマップ内の値を返す。無ければfloat.NaNを返す。</returns>
        public float GetValue(Pos _pos)
        {
            float ans = float.NaN;

            if (this.IsMapSet)                                              // マップがセットされていなければ、渡すデータが無い（それに、エラーが出る）
            {
                if (this.Useful(_pos))                                      // 指定座標が本マップに適用可能なら、値を抽出する
                {
                    ans = this.d_map[_pos.x, _pos.y];
                }
                else
                {
                    throw new SystemException("範囲外の座標が指定されました。");
                }
            }
            else
            {
                throw new SystemException("マップがセットされていません。初期化が必要です。");
            }
            return ans;
        }
        /// <summary>
        /// マップのサイズと角の座標からメッシュのサイズを計算して返す
        /// </summary>
        /// <param name="corner">角の座標値</param>
        /// <param name="mapSize">マップのサイズ</param>
        /// <returns>メッシュサイズ</returns>
        private Length GetMeshSize(RectangleField corner, Pos mapSize)
        {
            BLH _lower = corner.LowerLeft;
            BLH _upper = corner.UpperRight;
            Length _distance = _lower.GetDistance(_upper);                  // 2地点間の距離を求める。 10cm程度の誤差は無視しているので注意。
            return new Length(_distance.E / (double)mapSize.x, _distance.N / (double)mapSize.y); // メッシュのサイズを計算
        }
        /*2011/7/6 動作確認済み*/
        /// <summary>
        /// マップデータをセットする
        /// <para>データはディープコピーされます。</para>
        /// </summary>
        /// <param name="_map">マップのラスタデータ</param>
        public void SetMapData(float[,] _map)
        {
            Pos _size = new Pos(_map.GetLength(0), _map.GetLength(1));
            if (_size == new Pos()) throw new SystemException("マップのサイズが0です。");
            if (this.Size != _size)
            {
                //Console.WriteLine("MapBase.SetMapData(): マップを内部データに取り込む際に、セットマップサイズが異なるorセットされていなかったことを検出しました。引数の大きさに合わせてサイズを調整します。");
                this.Size = _size;                                          // サイズを代入
                this.MeshSize = this.GetMeshSize(this.Field, _size);        // メッシュサイズを計算して代入する
            }

            //this.d_map = _map;                                            // 実はシャローコピーでも処理時間はさほど変わらない。

            if (this.IsMapSet == false) this.d_map = new float[this.Size.x, this.Size.y]; // メモリ領域を確保
            // マップをセット           
            Parallel.For(0, this.Size.x, x =>                               // データをコピー
            {
                Parallel.For(0, this.Size.y, y=>
                {
                    this.d_map[x, y] = _map[x, y];
                });
            });
            return;
        }
        /// <summary>
        /// マップの初期化メソッド
        /// <para>
        /// マップサイズを定義済みであれば、マップの初期化を行います。
        /// マップサイズが定義済みかどうかはIsSizeSetメソッドを参照してご確認ください。
        /// なお、マップデータはデフォルトでは0.0fで上書きされます。
        /// </para>
        /// </summary>
        /// <param name="_value">初期化で埋めたい数値<para>省略すると0.0とします。</para></param>
        public void Initialize(float _value = 0.0f)
        {
            if (this.IsSizeSet)
            {
                if (this.IsMapSet == false) this.d_map = new float[this.Size.x, this.Size.y];   // メモリ領域を確保
                Parallel.For(0, this.Size.x, x =>
                {
                    for (int y = 0; y < this.Size.y; y++)
                    {
                        this.d_map[x, y] = _value;
                    }
                });
            }
            return;
        }
        #endregion
        #region "コンストラクタ関係"
        /// <summary>
        /// 共通した初期化項目
        /// <para>出力画像のフォーアットをビットマップに指定します。</para>
        /// </summary>
        private void CommonInit()
        {
            this.ImageFormat = System.Drawing.Imaging.ImageFormat.Bmp;
            return;
        }
        /*2011/7/6 動作確認済み*/
        /// <summary>
        /// マップの領域を指定してインスタンスを生成するコンストラクタ
        /// <para>
        /// 座標計算を特に使用しない場合を想定しています。
        /// マップのサイズはマップデータの代入時に決定されます。
        /// また、この時にマップデータのインスタンスが確保されます。
        /// </para>
        /// </summary>
        /// <param name="field">マップの領域情報</param>
        public Map(RectangleField field)
        {
            this.CommonInit();
            this.Size = new Pos();                                  // サイズの初期化だけはしておく
            this.Field = field;
            return;
        }
        /*2011/7/6 動作確認済み*/
        /// <summary>
        /// 領域の角の座標とマップのサイズ（ピクセル数）を指定することでマップを生成するコンストラクタ
        /// <para>
        /// マップは0で初期化されます。
        /// </para>
        /// </summary>
        /// <param name="field">マップの領域情報</param>
        /// <param name="_size">縦横のマップサイズ</param>
        public Map(RectangleField field, Pos _size)
        {
            this.CommonInit();
            this.Size = _size;
            this.MeshSize = this.GetMeshSize(field, this.Size);
            this.Field = field;                                     // 各種データを格納
            this.d_map = new float[this.Size.x, this.Size.y];       // マップの実態を確保
            return;
        }
        /*2011/7/6 動作確認済み*/
        /// <summary>
        /// 領域の角の座標とメッシュサイズを指定することでマップを生成するコンストラクタ
        /// <para>
        /// マップの境界値が曖昧だけども任意のマップを生成したいという場合を想定しています。
        /// なお、角の座標値はメッシュサイズに合わせて微調整されます。
        /// マップは0で初期化されます。
        /// </para>
        /// </summary>
        /// <param name="field">作りたいマップの領域情報（緯度・経度）</param>
        /// <param name="_meshSize">メッシュサイズ[m]</param>
        public Map(RectangleField field, Length _meshSize)
        {
            BLH _lower = field.LowerLeft;
            BLH _upper = field.UpperRight;
            Length _distance = _lower.GetDistance(_upper);                                                          // 2地点間の距離を求める。 高度の効果による10cm程度の誤差は無視しているので注意。
            Pos _size = new Pos((int)((Math.Abs(_distance.E) + 0.001) / _meshSize.E + 1.0d), (int)((Math.Abs(_distance.N) + 0.001) / _meshSize.N));      // マップのサイズを決める

            this.CommonInit();
            this.MeshSize = _meshSize;                                                                              // メッシュのサイズ（網目のサイズ==分解能）をコピー
            this.Size = _size;                                                                                      // マップのサイズをセット
            this.d_map = new float[this.Size.x, this.Size.y];
            _upper.L = _meshSize.E * (double)this.Size.x / _lower.GetUnitLengthForEN().E + field.LowerLeft.L;       // マップのサイズから、upperの座標値を修正する
            _upper.B = _meshSize.N * (double)this.Size.y / _lower.GetUnitLengthForEN().N + field.LowerLeft.B;       // 2地点中央の単位長でないのでやっぱり少しずれるけど、まあいいや
            this.Field = new RectangleField(_lower, _upper, AngleUnit.Degree, GNSS.GeodeticDatum.Datum.WGS84);      // メッシュ調整のためにupperだけ修正をかけたインスタンスを生成。
            return;
        }
        /// <summary>
        /// コピーコンストラクタ
        /// <para>パレメータを始め全てのパラメータをコピーしたインスタンスを生成します。</para>
        /// </summary>
        /// <param name="map">コピー元のマップ</param>
        public Map(Map map)
        {
            this.field = new RectangleField(map.field);
            this.size = map.size;
            this.errorSize = map.errorSize;
            this.meshSise = map.meshSise;
            this.imageFormat = map.imageFormat;

            if (map.d_map != null)
            {
                float[,] newMap = new float[map.d_map.GetLength(0), map.d_map.GetLength(1)];
                int x_width = map.Size.x;
                int y_height = map.Size.y;

                Parallel.For(0, x_width, x =>
                {
                    Parallel.For(0, y_height, y =>
                    {
                        newMap[x, y] = map.d_map[x, y];                 // 高速性を優先して、インデグサを用いない
                    });
                });
                this.d_map = newMap;
            }
            else
                this.d_map = null;
            return;
        }
        /// <summary>
        /// デスコンストラクタ
        /// </summary>
        ~Map()
        {
            this.d_map = null;
        }
        #endregion
    }
}
